<?php
namespace nulib\db\pgsql;

use nulib\cl;
use nulib\db\_private\_config;
use nulib\db\_private\Tvalues;
use nulib\db\IDatabase;
use nulib\db\ITransactor;
use nulib\php\func;

class Pgsql implements IDatabase {
  use Tvalues;

  static function with($pgsql, ?array $params=null): self {
    if ($pgsql instanceof static) {
      return $pgsql;
    } elseif ($pgsql instanceof self) {
      # recréer avec les mêmes paramètres
      return new static(null, cl::merge([
        "dbconn" => $pgsql->dbconn,
        "options" => $pgsql->options,
        "config" => $pgsql->config,
        "migrate" => $pgsql->migration,
      ], $params));
    } else {
      return new static($pgsql, $params);
    }
  }


  protected const OPTIONS = [
    "persistent" => true,
    "force_new" => false,
  ];

  const CONFIG = null;

  const MIGRATE = null;

  const params_SCHEMA = [
    "dbconn" => ["array"],
    "options" => ["?array|callable"],
    "replace_config" => ["?array|callable"],
    "config" => ["?array|callable"],
    "migrate" => ["?array|string|callable"],
    "auto_open" => ["bool", true],
  ];

  const dbconn_SCHEMA = [
    "" => "?string",
    "host" => "string",
    "hostaddr" => "?string",
    "port" => "?int",
    "dbname" => "string",
    "user" => "string",
    "password" => "string",
    "connect_timeout" => "?int",
    "options" => "?string",
    "sslmode" => "?string",
    "service" => "?string",
  ];

  protected const dbconn_MAP = [
    "name" => "dbname",
    "pass" => "password",
  ];

  const options_SCHEMA = [
    "persistent" => ["bool", self::OPTIONS["persistent"]],
    "force_new" => ["bool", self::OPTIONS["force_new"]],
  ];

  function __construct($dbconn=null, ?array $params=null) {
    if ($dbconn !== null) {
      if (!is_array($dbconn)) {
        $dbconn = ["" => $dbconn];
        #XXX à terme, il faudra interroger config
        #$tmp = config::db($dbconn);
        #if ($tmp !== null) $dbconn = $tmp;
        #else $dbconn = ["" => $dbconn];
      }
      $params["dbconn"] = $dbconn;
    }
    # dbconn
    $this->dbconn = $params["dbconn"] ?? null;
    # options
    $this->options = $params["options"] ?? static::OPTIONS;
    # configuration
    $config = $params["replace_config"] ?? null;
    if ($config === null) {
      $config = $params["config"] ?? static::CONFIG;
      if (is_callable($config)) $config = [$config];
    }
    $this->config = $config;
    # migrations
    $this->migration = $params["migrate"] ?? static::MIGRATE;
    #
    $defaultAutoOpen = self::params_SCHEMA["auto_open"][1];
    if ($params["auto_open"] ?? $defaultAutoOpen) {
      $this->open();
    }
  }

  protected ?array $dbconn;

  /** @var array|callable|null */
  protected $options;

  /** @var array|string|callable */
  protected $config;

  /** @var array|string|callable */
  protected $migration;

  /** @var resource */
  protected $db = null;

  function open(): self {
    if ($this->db === null) {
      $dbconn = $this->dbconn;
      $connection_string = [$dbconn[""] ?? null];
      unset($dbconn[""]);
      foreach ($dbconn as $key => $value) {
        if ($value === null) continue;
        $value = strval($value);
        if ($value === "" || preg_match("/[ '\\\\]/", $value)) {
          $value = str_replace("\\", "\\\\", $value);
          $value = str_replace("'", "\\'", $value);
          $value = "'$value'";
        }
        $key = cl::get(self::dbconn_MAP, $key, $key);
        $connection_string[] = "$key=$value";
      }
      $connection_string = implode(" ", array_filter($connection_string));
      $options = $this->options;
      if (is_callable($options)) {
        $options = func::with($options)->bind($this, true)->invoke();
      }
      $forceNew = $options["force_new"] ?? false;
      $flags = $forceNew? PGSQL_CONNECT_FORCE_NEW: 0;

      if ($options["persistent"] ?? true) $db = pg_pconnect($connection_string, $flags);
      else $db = pg_connect($connection_string, $flags);
      if ($db === false) throw new PgsqlException("unable to connect");
      $this->db = $db;

      _config::with($this->config)->configure($this);
      //_migration::with($this->migration)->migrate($this);
    }
    return $this;
  }

  function close(): self {
    if ($this->db !== null) {
      pg_close($this->db);
      $this->db = null;
    }
    return $this;
  }

  protected function db() {
    $this->open();
    return $this->db;
  }

  function exec($query, ?array $params = null) {
    // TODO: Implement exec() method.
  }

  function willUpdate(...$transactors): \nulib\db\ITransactor {
    // TODO: Implement willUpdate() method.
  }

  function inTransaction(): bool {
    // TODO: Implement inTransaction() method.
  }

  function beginTransaction(?callable $func = null, bool $commit = true): void {
    // TODO: Implement beginTransaction() method.
  }

  function commit(): void {
    // TODO: Implement commit() method.
  }

  function rollback(): void {
    // TODO: Implement rollback() method.
  }

  function get($query, ?array $params = null, bool $entireRow = false) {
    // TODO: Implement get() method.
  }

  function one($query, ?array $params = null): ?array {
    // TODO: Implement one() method.
  }

  function all($query, ?array $params = null, $primaryKeys = null): iterable {
    // TODO: Implement all() method.
  }
}
